{ ================================================= }
{ === unit MINIMISE                             === }
{ === version 1 du 2.12.95                      === }
{ === Frdric Ott                              === }
{ ================================================= }

unit minimise;

INTERFACE

uses math,wmagoptu;

const
    nbrMaxPara = 20; {nombre maximal de paramtres que l'on espere fitter }
type
    { type decrivant un vecteur de parametres ajustes }
    vecteurPara = array[1..nbrMaxPara] of double;

{ objet MINIMISATION                                            }
{ contient un ensemble de methodes tres generales de fcts       }
{ de minimisation 1D et MultiD                                  }
{ Pour l'utiliser sur une fonction F 1D ou multi D              }
{ il faut definir un type heritant Minimisation                 }
{ et definir une methode Fct1D ou FctMultiD pour la fct adhoc   }
{ ne pas oublier de lancer le constructor INIT avant utilisation}
{ EXEMPLE:                                                      }
{ type TypeMaFonctionQueJeVeuxMinimiser = object(minimisation)  }
{              function Fct1D(x:double):double;virtual;         }
{              end;                                             }
{ Function TypeMaFonctionQueJeVeuxMinimiser.Fct1D(x:double):double;}
{ begin                                                         }
{     Fct1D:=sin(x);                                            }
{ end;                                                          }
{ var f:TypeMaFonctionQueJeVeuxMinimiser                        }
{ begin                                                         }
{     f.init;  etc...                                           }
type
    Minimisation = object
             constructor init;
             { ensemble de procedures tres generales }
             Function FctMultiD(var point:VecteurPara):double;virtual;
             Function Fct1D(x:double):double; virtual; {ok}
             Procedure Bracketing(var a,b,c,fa,fb,fc:double); {ok}
             Procedure Golden(a,b,c,tol:double;var xmin,FMin:double); {ok}
             Procedure Brent(ax,bx,cx,tol:double;var xmin,FMin:double); {ok}
             procedure MultiD(N:integer;Iter:integer;tol:double;var Point:vecteurPara;var Value:double); {ok}
             end;

{type
    MinimisationMagOpt=object(minimisation)
             Function FctMultiD(var point:VecteurPara):double;virtual;
             Procedure MinimiserParaML(var L:MultiCouche;AjusterDansL:AjusterDansMultiCouche;
                              data:PdataArray;iter:integer;tol:double;var value:double);
             end; }

Constructor Minimisation.init;
begin
end;

{ Fonction multiD que l'on souhaite minimiser }
{ ENTREE: point, un vecteurPara               }
{ SORTIE: valeur de la fonction en ce point   }
Function Minimisation.FctMultiD(var point:VecteurPara):double;
begin
     FctMultiD:=Sin(Point[1]*Pi/180)*cos(Point[2]*Pi/180)*Sin(Point[3]*Pi/180);
end;

{ fonction 1D a minimiser                       }
{ cette fonction peut etre appelee par          }
{ MinBracketing, Minimisation1DGolden,          }
{ et Minimisation1DBrent                        }
{ ENTREE: un scalaire x                         }
{ SORTIE: la valeur de la fonction MultiD       }
Function Minimisation.Fct1D(x:double):double;
var j:integer;
    PointAux:VecteurPara;
begin
     Fct1D:=sin(x);
end;

{ routine de bracketing de minimum                      }
{ Numerical Recipes p.281                               }
{ ENTREE: a,b initial guesses                           }
{         puis on cherche un troisieme point            }
{ IMPLICITE: FctAMinimiser1D                            }
{ SORTIE: a,b,c qui encadrent un minimum                }
{         fa,fb,fc, valeurs de la fonction en a,b,c     }
{ NB: procedure semble ok                               }
{     et assez robuste                                  }
{ Bug:dans le cas ou il n'y a pas de minimum -> overflow}
Procedure Minimisation.Bracketing(var a,b,c,fa,fb,fc:double);
const
     Gold = 1.618034;
     GLimit=100;
     tiny=1e-20;
var
   Fin:boolean;
   r,q,u,uLim,fu,aux:double;
begin
     fa:=Fct1D(a);
     fb:=Fct1D(b);
     if fb>fa { on permute a et b pour suivre la pente descendante de a vers b }
     then
     begin
          aux:=a;a:=b;b:=aux;
          aux:=fa;fa:=fb;fb:=aux;
     end;
     c:=b+Gold*(b-a);
     fc:=Fct1D(c);

     Fin:=False;
     while Not(Fin) do
     begin
          if fb>=fc
          then
              begin
              r:=(b-a)*(fb-fc);
              q:=(b-c)*(fb-fa);
              u:=b-((b-c)*q-(b-a)*r)/(2*Max(abs(q-r),tiny)*signe(q-r));
              uLim:=b+GLimit*(c-b);
              if ((b-u)*(u-c))>0
              then begin
                   fu:=Fct1D(u);
                   if (fu<fc)
                   then begin
                        a:=b;
                        fa:=fb;
                        b:=u;
                        fb:=fu;
                        Fin:=True;
                        end
                   else if (fu>fb)
                        then begin
                             c:=u;fc:=fu;Fin:=True;
                             end;
                   u:=c+Gold*(c-b);
                   fu:=Fct1D(u);
                   end
              else begin
                   if ((c-u)*(u-uLim)>0)
                   then begin
                        fu:=Fct1D(u);
                        if (fu<fc)
                        then begin
                             b:=c;
                             c:=u;
                             u:=c+Gold*(c-b);
                             fb:=fc;
                             fc:=fu;
                             fu:=Fct1D(u);
                             end;
                        end
                   else if ((u-uLim)*(uLim-c)>=0)
                        then begin u:=uLim;fu:=Fct1D(u) end
                        else begin u:=c+Gold*(c-b);fu:=Fct1D(u) end;
                   end;
              if fin=false
              then
                  begin
                   a:=b;b:=c;c:=u;
                   fa:=fb;fb:=fc;fc:=fu;
                  end;
              end
          else Fin:=True;
     end;
end;


{ procedure de minimisation basee sur Golden section search }
{ (methode rustique mais robuste j'espere                   }
{ reference NUMERICAL RECIPES p.283                         }
{ ENTREE: a,b,c triplet encadrant un minimum                }
{         tol, precision du minimum                         }
{ implicite: la fonction a minimiser porte le nom           }
{            FctAMinimiser1D                                }
{ SORTIE: xmin abscisse du minimum                          }
{         fmin valeur de la fonction au minimum             }
{ procedure semble ok                                       }
Procedure Minimisation.Golden(a,b,c,tol:double;var xmin,FMin:double);
const
     NbrOr = 0.38196601125;
     UnMoinsNbrOr = 1 - NbrOr;
var
   fin:boolean;
   X0,X1,X2,X3,F0,F1,F2,F3:double;
begin
     X0:=a;
     X3:=c;
     if (abs(c-b)>abs(b-a))
     then begin X1:=b;X2:=b+NbrOr*(c-b) end
     else begin X2:=b;X1:=b-NbrOr*(b-a) end;

     F1:=Fct1D(X1);
     F2:=Fct1D(X2);

     fin:=False;
     while not(fin) do
     if (abs(X3-X0)>(tol*(abs(X1)+Abs(X2))))
     then begin
          if F2<F1
          then
              begin
              X0:=X1;X1:=X2;X2:=UnMoinsNbrOr*X1+NbrOr*X3;
              F0:=F1;F1:=F2;F2:=Fct1D(X2)
              end
          else
              begin
              X3:=X2;X2:=X1;X1:=UnMoinsNbrOr*X2+NbrOr*X0;
              F3:=F2;F2:=F1;F1:=Fct1D(X1)
              end
          end
     else fin:=true;


     If F1<F2
     then begin Fmin:=F1;Xmin:=X1 end
     else begin Fmin:=F2;Xmin:=X2 end;

end;

{ procedure de minimisation basee sur Brent                 }
{ implementation assez delicate (robustesse ?               }
{ reference NUMERICAL RECIPES p.284                         }
{ ENTREE: a,b,c triplet encadrant un minimum                }
{         tol, precision du minimum                         }
{ implicite: la fonction a minimiser porte le nom           }
{            FctAMinimiser1D                                }
{ SORTIE: xmin abscisse du minimum                          }
{         fmin valeur de la fonction au minimum             }
{ procedure semble ok                                       }
Procedure Minimisation.Brent(ax,bx,cx,tol:double;var xmin,FMin:double);
const
     NbrOr = 0.38196601125;
     UnMoinsNbrOr = 1 - NbrOr;
     Epsilon = 1e-10;
     IterMax = 100;
var
   i:integer;
   FinMin:boolean;
   a,b,c,d,u,v,w,x,xm,E,Etemp,Tol1,Tol2,p,q,r:double;
   Fx,Fu,Fv,Fw:double;
   value:double;
begin
     { initialisations }
     { a et b doivent etre dans l'ordre croissant }
     a:=Min(ax,cx);
     b:=Max(ax,cx);
     v:=bx;w:=v;x:=v;
     E:=0;
     Fx:=Fct1D(x);
     Fv:=Fx;
     Fw:=Fx;

     i:=0;
     FinMin:=False;
     while (i<=IterMax) and Not(FinMin) do
     begin
          Xm:=0.5*(a+b);
          Tol1:=Tol*abs(x)+Epsilon;
          Tol2:=2*Tol1;
          if (abs(x-xm)<=(Tol2-0.5*(b-a)))
          then FinMin:=True
          else
          begin
          If (abs(E)>Tol1)
          then begin
               r:=(x-w)*(fx-fv);
               q:=(x-v)*(fx-fw);
               p:=(x-v)*q-(x-w)*R;
               q:=2*(q-r);
               If (q>0) then p:=-p;
               q:=abs(q);
               Etemp:=E;
               E:=d;
               If (Abs(p)>=Abs(0.5*q*Etemp))
               or (P<=(q*(a-x)))
               or (P>=(q*(b-x)))
               then begin
                    if (x>=xm) then e:=a-x else e:=b-x;
                    d:=NbrOr*e
                    end
               else begin
                    d:=p/q;
                    u:=x+d;
                    if ((u-a)<tol2) or ((b-u)<tol2) then d:=Tol1*Signe(xm-x);
                    end
               end
          else begin
               if (x>=xm) then e:=a-x else e:=b-x;
               d:=NbrOr*e
               end;

          If abs(d)>=Tol1
          then u:=x+d
          else u:=x+tol1*signe(d);
          Fu:=Fct1D(u);
          If (Fu<=Fx)
          then begin
               if u>=x then a:=x else b:=x;
               v:=w;fv:=fw;
               w:=x;fw:=fx;
               x:=u;fx:=fu;
               end
          else begin
               if u<x then a:=u else b:=u;
               if (Fu<=fW) or (w=x)
               then begin v:=w;Fv:=Fw;w:=u;Fw:=Fu end
               else if (Fu<=Fv) or (v=x) or (v=w)
                    then begin v:=u;Fv:=Fu end;
               end
          end;
     end;

     if i>IterMax
     then writeln('Nombre maximal d''iterations depasse dans Minimisation Brent');

     Xmin:=x;
     Fmin:=Fx;

end;


{ ======================================================== }
{ =====  CAS MULTIDIMENTIONNEL   ========================= }
{ =====  les procedures prcdntes sont reintegrees ====== }
{ =====  dans la procedure MultiD  ======================= }

{ procedure de minimisation base sur la mthode de Powell }
{ Rfrence NUMERICAL RECIPES p.298                        }
{ ENTREE: Point, starting point                            }
{         N, nombre de variables a minimiser               }
{         Iter nombre d'itrations demandes               }
{ SORTIE: Point, le vecteur de minimisation                }
{         Value, la valeur de la fonction                  }
{ NB: la fonction a minimiser porte le nom FctAMinimiserMultiD }
{ il faut la rentrer a la main dans cette unit ...            }
{ les points de depart semble influer pas mal                  }
{ il y a d'ailleurs risque d'overflow ou invalid float         }
procedure Minimisation.MultiD(N:integer;Iter:integer;tol:double;var Point:vecteurPara;var Value:double);
var
     i,j,k:integer;
     FinMultiMin:boolean;
     IndiceBiggestDecrease:integer;
     BiggestDecrease:Double;
     PointAux1,PointAux2:vecteurPara;
     ValueAux1,ValueAux2:double;
     DirectionVect:vecteurPara;
     DirectionMat:array[1..NbrMaxPara,1..NbrMaxPara] of double; { ensemble des directions de minimisation }

{***************************************************************}
{ Evaluation de F(P+l*n)                        }
{ ENTREE: un scalaire x                         }
{        IMPLICITE: un point P et une           }
{        direction DirectionVect                }
{ SORTIE: la valeur de la fonction MultiD       }
{         au point P + x * Direction            }
Function FctAMinimiser1D(x:double):double;
var j:integer;
    PointAux:VecteurPara;
begin
     for j:=1 to N do PointAux[j]:=Point[j]+DirectionVect[j]*x;
     FctAMinimiser1D:=FctMultiD(PointAux);
end;

{ ****************************************************************}

{ routine de bracketing de minimum                      }
{ Numerical Recipes p.281                               }
{ ENTREE: a,b initial guesses                           }
{         puis on cherche un troisieme point            }
{ IMPLICITE: FctAMinimiser1D                            }
{ SORTIE: a,b,c qui encadrent un minimum                }
{         fa,fb,fc, valeurs de la fonction en a,b,c     }
{ NB: procedure semble ok                               }
{     et assez robuste                                  }
{ Bug:dans le cas ou il n'y a pas de minimum -> overflow}
Procedure MinBracketing(var a,b,c,fa,fb,fc:double);
const
     Gold = 1.618034;
     GLimit=100;
     tiny=1e-20;
var
   Fin:boolean;
   r,q,u,uLim,fu,aux:double;
begin
     fa:=FctAMinimiser1D(a);
     fb:=FctAMinimiser1D(b);
     if fb>fa { on permute a et b pour suivre la pente descendante de a vers b }
     then
     begin
          aux:=a;a:=b;b:=aux;
          aux:=fa;fa:=fb;fb:=aux;
     end;
     c:=b+Gold*(b-a);
     fc:=FctAMinimiser1D(c);

     Fin:=False;
     while Not(Fin) do
     begin
          if fb>=fc
          then
              begin
              r:=(b-a)*(fb-fc);
              q:=(b-c)*(fb-fa);
              u:=b-((b-c)*q-(b-a)*r)/(2*Max(abs(q-r),tiny)*signe(q-r));
              uLim:=b+GLimit*(c-b);
              if ((b-u)*(u-c))>0
              then begin
                   fu:=FctAMinimiser1D(u);
                   if (fu<fc)
                   then begin
                        a:=b;
                        fa:=fb;
                        b:=u;
                        fb:=fu;
                        Fin:=True;
                        end
                   else if (fu>fb)
                        then begin
                             c:=u;fc:=fu;Fin:=True;
                             end;
                   u:=c+Gold*(c-b);
                   fu:=FctAMinimiser1D(u);
                   end
              else begin
                   if ((c-u)*(u-uLim)>0)
                   then begin
                        fu:=FctAMinimiser1D(u);
                        if (fu<fc)
                        then begin
                             b:=c;
                             c:=u;
                             u:=c+Gold*(c-b);
                             fb:=fc;
                             fc:=fu;
                             fu:=FctAMinimiser1D(u);
                             end;
                        end
                   else if ((u-uLim)*(uLim-c)>=0)
                        then begin u:=uLim;fu:=FctAMinimiser1D(u) end
                        else begin u:=c+Gold*(c-b);fu:=FctAMinimiser1D(u) end;
                   end;
              if fin=false
              then
                  begin
                   a:=b;b:=c;c:=u;
                   fa:=fb;fb:=fc;fc:=fu;
                  end;
              end
          else Fin:=True;
     end;

end;

{ procedure de minimisation basee sur Golden section search }
{ (methode rustique mais robuste j'espere                   }
{ reference NUMERICAL RECIPES p.283                         }
{ ENTREE: a,b,c triplet encadrant un minimum                }
{         tol, precision du minimum                         }
{ implicite: la fonction a minimiser porte le nom           }
{            FctAMinimiser1D                                }
{ SORTIE: xmin abscisse du minimum                          }
{         fmin valeur de la fonction au minimum             }
{ procedure semble ok                                       }
Procedure Minimisation1DGolden(a,b,c,tol:double;var xmin,FMin:double);
const
     NbrOr = 0.38196601125;
     UnMoinsNbrOr = 1 - NbrOr;
var
   fin:boolean;
   X0,X1,X2,X3,F0,F1,F2,F3:double;
begin
     X0:=a;
     X3:=c;
     if (abs(c-b)>abs(b-a))
     then begin X1:=b;X2:=b+NbrOr*(c-b) end
     else begin X2:=b;X1:=b-NbrOr*(b-a) end;

     F1:=FctAMinimiser1D(X1);
     F2:=FctAMinimiser1D(X2);

     fin:=False;
     while not(fin) do
     if (abs(X3-X0)>(tol*(abs(X1)+Abs(X2))))
     then begin
          if F2<F1
          then
              begin
              X0:=X1;X1:=X2;X2:=UnMoinsNbrOr*X1+NbrOr*X3;
              F0:=F1;F1:=F2;F2:=FctAMinimiser1D(X2)
              end
          else
              begin
              X3:=X2;X2:=X1;X1:=UnMoinsNbrOr*X2+NbrOr*X0;
              F3:=F2;F2:=F1;F1:=FctAMinimiser1D(X1)
              end
          end
     else fin:=true;


     If F1<F2
     then begin Fmin:=F1;Xmin:=X1 end
     else begin Fmin:=F2;Xmin:=X2 end;

end;


{ procedure de minimisation basee sur Brent                 }
{ implementation assez delicate (robustesse ?               }
{ reference NUMERICAL RECIPES p.284                         }
{ ENTREE: a,b,c triplet encadrant un minimum                }
{         tol, precision du minimum                         }
{ implicite: la fonction a minimiser porte le nom           }
{            FctAMinimiser1D                                }
{ SORTIE: xmin abscisse du minimum                          }
{         fmin valeur de la fonction au minimum             }
{ procedure semble ok                                       }
Procedure Minimisation1DBrent(ax,bx,cx,tol:double;var xmin,FMin:double);
const
     NbrOr = 0.38196601125;
     UnMoinsNbrOr = 1 - NbrOr;
     Epsilon = 1e-10;
     IterMax = 100;
var
   i:integer;
   FinMin:boolean;
   a,b,c,d,u,v,w,x,xm,E,Etemp,Tol1,Tol2,p,q,r:double;
   Fx,Fu,Fv,Fw:double;
   value:double;
begin
     { initialisations }
     { a et b doivent etre dans l'ordre croissant }
     a:=Min(ax,cx);
     b:=Max(ax,cx);
     v:=bx;w:=v;x:=v;
     E:=0;
     Fx:=FctAMinimiser1D(x);
     Fv:=Fx;
     Fw:=Fx;

     i:=0;
     FinMin:=False;
     while (i<=IterMax) and Not(FinMin) do
     begin
          Xm:=0.5*(a+b);
          Tol1:=Tol*abs(x)+Epsilon;
          Tol2:=2*Tol1;
          if (abs(x-xm)<=(Tol2-0.5*(b-a)))
          then FinMin:=True
          else
          begin
          If (abs(E)>Tol1)
          then begin
               r:=(x-w)*(fx-fv);
               q:=(x-v)*(fx-fw);
               p:=(x-v)*q-(x-w)*R;
               q:=2*(q-r);
               If (q>0) then p:=-p;
               q:=abs(q);
               Etemp:=E;
               E:=d;
               If (Abs(p)>=Abs(0.5*q*Etemp))
               or (P<=(q*(a-x)))
               or (P>=(q*(b-x)))
               then begin
                    if (x>=xm) then e:=a-x else e:=b-x;
                    d:=NbrOr*e
                    end
               else begin
                    d:=p/q;
                    u:=x+d;
                    if ((u-a)<tol2) or ((b-u)<tol2) then d:=Tol1*Signe(xm-x);
                    end
               end
          else begin
               if (x>=xm) then e:=a-x else e:=b-x;
               d:=NbrOr*e
               end;

          If abs(d)>=Tol1
          then u:=x+d
          else u:=x+tol1*signe(d);
          Fu:=FctAMinimiser1D(u);
          If (Fu<=Fx)
          then begin
               if u>=x then a:=x else b:=x;
               v:=w;fv:=fw;
               w:=x;fw:=fx;
               x:=u;fx:=fu;
               end
          else begin
               if u<x then a:=u else b:=u;
               if (Fu<=fW) or (w=x)
               then begin v:=w;Fv:=Fw;w:=u;Fw:=Fu end
               else if (Fu<=Fv) or (v=x) or (v=w)
                    then begin v:=u;Fv:=Fu end;
               end
          end;
     end;

     if i>IterMax
     then writeln('Nombre maximal d''iterations depasse dans Minimisation Brent');

     Xmin:=x;
     Fmin:=Fx;

end;



{ Minimisation dans la direction Point+l*DirectionVect }
Procedure LineMinimization(var Point:vecteurPara;var DirectionVect:vecteurPara;N:integer;var Value:double);
var
   j:integer;
   Xmin:double;
   a,x,b,fa,fx,fb:double;
begin
     a:=0;x:=1;b:=2;
     MinBracketing(a,x,b,fa,fx,fb);
     Minimisation1DGolden(a,x,b,tol,Xmin,Value);

     For j:=1 to N do
     begin
          DirectionVect[j]:=Xmin*DirectionVect[j];
          Point[j]:=DirectionVect[j]+Point[j];
     end;
end;

begin
     { initialisation du vecteur de directions aux N vecteurs unitaires }
     for i:=1 to N do
         for j:=1 to N do
              begin
              DirectionMat[i,j]:=0;
              if i=j then DirectionMat[i,j]:=1
              end;

     Value:=FctMultiD(Point);
     { sauvegarde du point initial }
     PointAux1:=Point;

     FinMultiMin:=False;
     k:=0;
     { boucle de Iter itrations }
     While ((k<=Iter) and Not(FinMultiMin)) do
     begin
          k:=k+1;
          ValueAux1:=Value;
          IndiceBiggestDecrease:=0;
          BiggestDecrease:=0;
          for i:=1 to N do
          begin
              for j:=1 to N do DirectionVect[j]:=DirectionMat[j,i];
              ValueAux2:=Value;
              LineMinimization(Point,DirectionVect,N,Value); {minimisation le long du vecteur DirectionVect }
              If (abs(ValueAux2-Value)>BiggestDecrease)
              then begin
                   BiggestDecrease:=abs(ValueAux2-Value);
                   IndiceBiggestDecrease:=i;
                   end;
          end;

          { termination criterion }
          if 2*abs(ValueAux1-Value)<=(tol*(abs(ValueAux1)+abs(Value)))
          then finMultiMin:=true;

          for j:=1 to N do
          begin
               PointAux2[j]:=2*Point[j]-PointAux1[j];
               DirectionVect[j]:=Point[j]-PointAux1[j];
               PointAux1[j]:=Point[j];
          end;

          valueAux2:=FctMultiD(PointAux2);
          if (ValueAux2>=ValueAux1) or
             ((2*(ValueAux1-2*Value+ValueAux2)*sqr(ValueAux1-Value-BiggestDecrease)
             -BiggestDecrease*sqr(ValueAux1-ValueAux2))>=0)
          then {on ne change pas de direction }
          else begin
               LineMinimization(Point,DirectionVect,N,Value);
               for j:=1 to N do DirectionMat[j,IndiceBiggestDecrease]:=DirectionVect[j];
               end;
     end; {fin de la boucle en k }

end;

{ fin procedure minimisation multi D }


Type TaMinimiser=object(Minimisation)
                    Function Fct1D(x:double):double; virtual;
                    Function FctMultiD(var point:VecteurPara):double;virtual;
                    end;
Function TaMinimiser.Fct1D(x:double):double;
begin
     Fct1D:=Cos(x*Pi/180);
end;

Function TaMinimiser.FctMultiD(var point:VecteurPara):double;
begin
     FctMultiD:=sin(Point[1]*Pi/180)*cos(Point[2]*Pi/180)*sin(Point[3]*Pi/180);
end;

var aMinimiser:Taminimiser;

var a,b,c,x,fa,fb,fc,fx:double;
    point:vecteurPara;

begin
     clrscr;
     aMinimiser.init;
     a:=10;b:=15;c:=20;
     Point[1]:=0;Point[2]:=0;Point[3]:=1;
     aMinimiser.Bracketing(a,b,c,fa,fb,fc);
     aMinimiser.Golden(a,b,c,0.001,x,fx);
     aMinimiser.MultiD(3,10,0.0001,Point,fx);
     writeln('x = ',Point[1]);
     writeln('x = ',Point[2]);
     writeln('x = ',Point[3]);
     writeln('f = ',fx);
end.


{ ============================================================================================= }
{                        cas de la simulation magneto-optique des multicouches                                  }
{ ============================================================================================= }


{ procedure de minimisation des fits de reflexion optique                                 }
{ ENTREE: une multicouche L, les parametres a ajuster dans cette multicouche AjusterDansL }
{         un pointeur vers des data experimentales  data                                  }
{         un nombre d'iteration Iter et une tolerance tol                                 }
{ SORTIE: ecart (moindres carres) entre les mesures et le fit: value                      }
{ MODIFIES: L est modifie et contient la couche ajustee                                  }
Procedure MinimiserParaML(var L:MultiCouche;AjusterDansL:AjusterDansMultiCouche;
                              data:PdataArray;iter:integer;tol:double;var value:double);
var
   i,j,N:integer;
   P:vecteurPara;

{ transfert les valeurs ajustables d'une multicouche vers un VecteurPara P }
{ ENTREE: L, multiCouche                                                   }
{ SORTIE: P, vecteurPara; n, nombre de parametres a ajuster                }
Procedure TransfertLversP(Var L:MultiCouche;AjusterDansL:AjusterDansMultiCouche;Var Point:VecteurPara;Var n:integer);
var i,j:integer;
    aux:complex;
begin
     { creation du vecteur a entrer dans la fct de minimisation }
     { L -> Point }
     j:=1;
     for i:=1 to nbrCouches do
         begin
              if AjusterDansL[i].epaisseur=True then begin Point[j]:=L[i].d;j:=j+1 end;
              if AjusterDansL[i].nii=True
              then if L[i].isotrope=true
                   then begin SqrtComp(L[i].EpsRelScal,aux);Point[j]:=Real(aux);j:=j+1 end
                   else begin SqrtComp(L[i].EpsRelMatPhys.c[1,1],aux);Point[j]:=Real(aux);j:=j+1 end;
              if AjusterDansL[i].kii=True
              then if L[i].isotrope=true
                   then begin SqrtComp(L[i].EpsRelScal,aux);Point[j]:=Im(aux);j:=j+1 end
                   else begin SqrtComp(L[i].EpsRelMatPhys.c[1,1],aux);Point[j]:=Im(aux);j:=j+1 end;
              if AjusterDansL[i].nij=True
              then begin SqrtComp(L[i].EpsRelMatPhys.c[1,2],aux);Point[j]:=Real(aux);j:=j+1 end;
              if AjusterDansL[i].kij=True
              then begin SqrtComp(L[i].EpsRelMatPhys.c[1,2],aux);Point[j]:=Im(aux);j:=j+1 end;
              if AjusterDansL[i].theta=True then begin Point[j]:=L[i].theta;j:=j+1 end;
              if AjusterDansL[i].phi=True then begin Point[j]:=L[i].phi;j:=j+1 end;
         end;
     n:=j-1;
end;

{ transfert les valeurs ajustes du vecteurPara vers la mutliCouche L }
{ ENTREE: P, vecteurPara                                              }
{ SORTIE: L, multicouche                                              }
{ un peu lourd tous ces transferts ... }
procedure TransfertPversL(Var Point:VecteurPara;var L:MultiCouche;AjusterDansL:AjusterDansMultiCouche);
var i,j,k:integer;
    aux,aux1:complex;
begin
 { transfert des parametres du point vers L }
     { Point -> L }
     j:=1;
     for i:=1 to nbrCouches do
         begin
              if AjusterDansL[i].epaisseur=True then begin L[i].d:=Point[j];j:=j+1 end;
              if AjusterDansL[i].nii=True
              then if L[i].isotrope=true
                   then begin
                        aux.x:=Point[j];j:=j+1;
                        SqrtComp(L[i].EpsRelScal,aux1);
                        aux.y:=Im(aux1);
                        MultComp(aux,aux,L[i].EpsRelscal)
                        end
                   else begin
                        aux.x:=Point[j];j:=j+1;
                        SqrtComp(L[i].EpsRelMatPhys.c[1,1],aux1);
                        aux.y:=Im(aux1);
                        MultComp(aux,aux,aux1);
                        for k:=1 to 3 do L[i].EpsRelMatPhys.c[j,j]:=aux1;
                        end;
              if AjusterDansL[i].kii=True
              then if L[i].isotrope=true
                   then begin
                        aux.y:=Point[j];j:=j+1;
                        SqrtComp(L[i].EpsRelScal,aux1);
                        aux.x:=Real(aux1);
                        MultComp(aux,aux,L[i].EpsRelscal)
                        end
                   else begin
                        aux.y:=Point[j];j:=j+1;
                        SqrtComp(L[i].EpsRelMatPhys.c[1,1],aux1);
                        aux.x:=Im(aux1);
                        MultComp(aux,aux,aux1);
                        for k:=1 to 3 do L[i].EpsRelMatPhys.c[j,j]:=aux1;
                        end;
              if AjusterDansL[i].nij=True
              then begin
                        aux.x:=Point[j];j:=j+1;
                        SqrtComp(L[i].EpsRelMatPhys.c[1,2],aux1);
                        aux.y:=Im(aux1);
                        MultComp(aux,aux,aux1);
                        L[i].EpsRelMatPhys.c[1,2]:=aux1;
                        NegComp(aux,L[i].EpsRelMatPhys.c[2,1]);
                        end;
              if AjusterDansL[i].kij=True
              then begin
                        aux.y:=Point[j];j:=j+1;
                        SqrtComp(L[i].EpsRelMatPhys.c[1,2],aux1);
                        aux.x:=Im(aux1);
                        MultComp(aux,aux,aux1);
                        L[i].EpsRelMatPhys.c[1,2]:=aux1;
                        NegComp(aux,L[i].EpsRelMatPhys.c[2,1]);
                        end;
              if AjusterDansL[i].theta=True then begin L[i].theta:=Point[j];j:=j+1 end;
              if AjusterDansL[i].phi=True then begin L[i].phi:=Point[j];j:=j+1 end;
         end;
     MiseAJourTenseursPermittivite(L);
end;


{ Procedure calculant l'ecart (moindres carres) entre les data experimentales }
{ et la couche propose                                                       }
{ ENTREE: L la multicouche consideree                                         }
{         data, les data experimentales                                       }
{ IMPLICITE: OrdreCouches                                                     }
{ SORTIE: l'ecart (moindres carres entre les deux                             }
Function CalculerEcart(var L:MultiCouche;var data:PdataArray):double;
var i:integer;
    Psi,Delta,Theta,Gamma:double;
    Einc,Eref:vecteur; { vecteur champ lectrique incident et reflechi }
    R:matrice;
    Ecart:double;
begin
     InitVect(Einc,2);
     InitVect(Eref,2);
     { initialisation du vecteur d'onde incident }
     { cf. Penissard p.155 }
     Einc.c[1].x:=cos(OndeIncidente.theta*Pi/180)*cos(OndeIncidente.gamma*Pi/180);
     Einc.c[1].y:=-sin(OndeIncidente.Theta*Pi/180)*sin(OndeIncidente.gamma*Pi/180);
     Einc.c[2].x:=sin(OndeIncidente.theta*Pi/180)*cos(OndeIncidente.gamma*Pi/180);
     Einc.c[2].y:=cos(OndeIncidente.Theta*Pi/180)*sin(OndeIncidente.gamma*Pi/180);

     ecart:=0;
     for i:=1 to data^.NbrLignes do
     begin
          AngleIncidence:=data^.data[i,1];
          CalculerR(L,k0,AngleIncidence,R);
          MultMatVect(R,Einc,Eref); { vecteur champ electrique reflechi }
          DeterminerParaEllipsometriques(Eref,Psi,Delta,Theta,Gamma);
          ecart:=Ecart+Sqr(data^.data[i,2]-Theta);
     end;
     CalculerEcart:=Ecart;
end;



{ ======================================================== }
{ =====  CAS MULTIDIMENTIONNEL   ========================= }
{ procedure de minimisation base sur la mthode de Powell }
{ Rfrence NUMERICAL RECIPES p.298                        }
{ procedure interne a MinimiserParaML                      }
{ ENTREE: Point, starting point                            }
{         N, nombre de variables a minimiser               }
{         Iter nombre d'itrations demandes               }
{ SORTIE: Point, le vecteur de minimisation                }
{         Value, la valeur de la fonction                  }
{ NB: la fonction a minimiser porte le nom FctAMinimiserMultiD }
{ il faut la rentrer a la main dans cette unit ...            }
procedure MinimisationMultiD(N:integer;Iter:integer;tol:double;var Point:vecteurPara;var Value:double);
var
     i,j,k:integer;
     FinMultiMin:boolean;
     IndiceBiggestDecrease:integer;
     BiggestDecrease:Double;
     PointAux1,PointAux2:vecteurPara;
     ValueAux1,ValueAux2:double;
     DirectionVect:vecteurPara;
     DirectionMat:array[1..NbrMaxPara,1..NbrMaxPara] of double; { ensemble des directions de minimisation }


{***************************************************************}
{ Fonction multiD que l'on souhaite minimiser }
{ ENTREE: point, un vecteurPara               }
{ SORTIE: valeur de la fonction en ce point   }
Function FctAMinimiserMultiD(var point:VecteurPara):double;
begin
     TransfertPversL(point,L,AjusterDansL);
     FctAMinimiserMultiD:=CalculerEcart(L,data);
end;
{***************************************************************}
{ Evaluation de F(P+l*n)                        }
{ ENTREE: un scalaire x                         }
{        IMPLICITE: un point P et une           }
{        direction DirectionVect                }
{ SORTIE: la valeur de la fonction MultiD       }
{         au point P + x * Direction            }
Function FctAMinimiser1D(x:double):double;
var j:integer;
    PointAux:VecteurPara;
begin
     for j:=1 to N do PointAux[j]:=Point[j]+DirectionVect[j]*x;
     FctAMinimiser1D:=FctAMinimiserMultiD(PointAux);
end;
{ ****************************************************************}

{ routine de bracketing de minimum                      }
{ Numerical Recipes p.281                               }
{ ENTREE: a,b initial guesses                           }
{         puis on cherche un troisieme point            }
{ IMPLICITE: FctAMinimiser1D                            }
{ SORTIE: a,b,c qui encadrent un minimum                }
{         fa,fb,fc, valeurs de la fonction en a,b,c     }
{ NB: procedure semble ok                               }
{     et assez robuste                                  }
{ Bug:dans le cas ou il n'y a pas de minimum -> overflow}
Procedure MinBracketing(var a,b,c,fa,fb,fc:double);
const
     Gold = 1.618034;
     GLimit=100;
     tiny=1e-20;
var
   Fin:boolean;
   r,q,u,uLim,fu,aux:double;
begin
     fa:=FctAMinimiser1D(a);
     fb:=FctAMinimiser1D(b);
     if fb>fa { on permute a et b pour suivre la pente descendante de a vers b }
     then
     begin
          aux:=a;a:=b;b:=aux;
          aux:=fa;fa:=fb;fb:=aux;
     end;
     c:=b+Gold*(b-a);
     fc:=FctAMinimiser1D(c);

     Fin:=False;
     while Not(Fin) do
     begin
          if fb>=fc
          then
              begin
              r:=(b-a)*(fb-fc);
              q:=(b-c)*(fb-fa);
              u:=b-((b-c)*q-(b-a)*r)/(2*Max(abs(q-r),tiny)*signe(q-r));
              uLim:=b+GLimit*(c-b);
              if ((b-u)*(u-c))>0
              then begin
                   fu:=FctAMinimiser1D(u);
                   if (fu<fc)
                   then begin
                        a:=b;
                        fa:=fb;
                        b:=u;
                        fb:=fu;
                        Fin:=True;
                        end
                   else if (fu>fb)
                        then begin
                             c:=u;fc:=fu;Fin:=True;
                             end;
                   u:=c+Gold*(c-b);
                   fu:=FctAMinimiser1D(u);
                   end
              else begin
                   if ((c-u)*(u-uLim)>0)
                   then begin
                        fu:=FctAMinimiser1D(u);
                        if (fu<fc)
                        then begin
                             b:=c;
                             c:=u;
                             u:=c+Gold*(c-b);
                             fb:=fc;
                             fc:=fu;
                             fu:=FctAMinimiser1D(u);
                             end;
                        end
                   else if ((u-uLim)*(uLim-c)>=0)
                        then begin u:=uLim;fu:=FctAMinimiser1D(u) end
                        else begin u:=c+Gold*(c-b);fu:=FctAMinimiser1D(u) end;
                   end;
              if fin=false
              then
                  begin
                   a:=b;b:=c;c:=u;
                   fa:=fb;fb:=fc;fc:=fu;
                  end;
              end
          else Fin:=True;
     end;

end;

{ procedure de minimisation basee sur Brent                 }
{ implementation assez delicate (robustesse ?               }
{ reference NUMERICAL RECIPES p.284                         }
{ ENTREE: a,b,c triplet encadrant un minimum                }
{         tol, precision du minimum                         }
{ implicite: la fonction a minimiser porte le nom           }
{            FctAMinimiser1D                                }
{ SORTIE: xmin abscisse du minimum                          }
{         fmin valeur de la fonction au minimum             }
{ procedure semble ok                                       }
Procedure Minimisation1DBrent(ax,bx,cx,tol:double;var xmin,FMin:double);
const
     NbrOr = 0.38196601125;
     UnMoinsNbrOr = 1 - NbrOr;
     Epsilon = 1e-10;
     IterMax = 100;
var
   i:integer;
   FinMin:boolean;
   a,b,c,d,u,v,w,x,xm,E,Etemp,Tol1,Tol2,p,q,r:double;
   Fx,Fu,Fv,Fw:double;
   value:double;
begin
     { initialisations }
     { a et b doivent etre dans l'ordre croissant }
     a:=Min(ax,cx);
     b:=Max(ax,cx);
     v:=bx;w:=v;x:=v;
     E:=0;
     Fx:=FctAMinimiser1D(x);
     Fv:=Fx;
     Fw:=Fx;

     i:=0;
     FinMin:=False;
     while (i<=IterMax) and Not(FinMin) do
     begin
          Xm:=0.5*(a+b);
          Tol1:=Tol*abs(x)+Epsilon;
          Tol2:=2*Tol1;
          if (abs(x-xm)<=(Tol2-0.5*(b-a)))
          then FinMin:=True
          else
          begin
          If (abs(E)>Tol1)
          then begin
               r:=(x-w)*(fx-fv);
               q:=(x-v)*(fx-fw);
               p:=(x-v)*q-(x-w)*R;
               q:=2*(q-r);
               If (q>0) then p:=-p;
               q:=abs(q);
               Etemp:=E;
               E:=d;
               If (Abs(p)>=Abs(0.5*q*Etemp))
               or (P<=(q*(a-x)))
               or (P>=(q*(b-x)))
               then begin
                    if (x>=xm) then e:=a-x else e:=b-x;
                    d:=NbrOr*e
                    end
               else begin
                    d:=p/q;
                    u:=x+d;
                    if ((u-a)<tol2) or ((b-u)<tol2) then d:=Tol1*Signe(xm-x);
                    end
               end
          else begin
               if (x>=xm) then e:=a-x else e:=b-x;
               d:=NbrOr*e
               end;

          If abs(d)>=Tol1
          then u:=x+d
          else u:=x+tol1*signe(d);
          Fu:=FctAMinimiser1D(u);
          If (Fu<=Fx)
          then begin
               if u>=x then a:=x else b:=x;
               v:=w;fv:=fw;
               w:=x;fw:=fx;
               x:=u;fx:=fu;
               end
          else begin
               if u<x then a:=u else b:=u;
               if (Fu<=fW) or (w=x)
               then begin v:=w;Fv:=Fw;w:=u;Fw:=Fu end
               else if (Fu<=Fv) or (v=x) or (v=w)
                    then begin v:=u;Fv:=Fu end;
               end
          end;
     end;

     if i>IterMax
     then writeln('Nombre maximal d''iterations depasse dans Minimisation Brent');

     Xmin:=x;
     Fmin:=Fx;

end;



{ Minimisation dans la direction Point+l*DirectionVect }
Procedure LineMinimization(var Point:vecteurPara;var DirectionVect:vecteurPara;N:integer;var Value:double);
var
   j:integer;
   Xmin:double;
   a,x,b,fa,fx,fb:double;
begin
     a:=0;x:=1;b:=2;
     MinBracketing(a,x,b,fa,fx,fb);
     Minimisation1DBrent(a,x,b,tol,Xmin,Value);

     For j:=1 to N do
     begin
          DirectionVect[j]:=Xmin*DirectionVect[j];
          Point[j]:=DirectionVect[j]+Point[j];
     end;
end;

begin
     { initialisation du vecteur de directions aux N vecteurs unitaires }
     for i:=1 to N do
         for j:=1 to N do
              begin
              DirectionMat[i,j]:=0;
              if i=j then DirectionMat[i,j]:=1
              end;

     Value:=FctAMinimiserMultiD(Point);
     { sauvegarde du point initial }
     PointAux1:=Point;

     FinMultiMin:=False;
     k:=0;
     { boucle de Iter itrations }
     While ((k<=Iter) and Not(FinMultiMin)) do
     begin
          k:=k+1;
          ValueAux1:=Value;
          IndiceBiggestDecrease:=0;
          BiggestDecrease:=0;
          for i:=1 to N do
          begin
              for j:=1 to N do DirectionVect[j]:=DirectionMat[j,i];
              ValueAux2:=Value;
              LineMinimization(Point,DirectionVect,N,Value); {minimisation le long du vecteur DirectionVect }
              If (abs(ValueAux2-Value)>BiggestDecrease)
              then begin
                   BiggestDecrease:=abs(ValueAux2-Value);
                   IndiceBiggestDecrease:=i;
                   end;
          end;

          { termination criterion }
          if 2*abs(ValueAux1-Value)<=(tol*(abs(ValueAux1)+abs(Value)))
          then finMultiMin:=true;

          for j:=1 to N do
          begin
               PointAux2[j]:=2*Point[j]-PointAux1[j];
               DirectionVect[j]:=Point[j]-PointAux1[j];
               PointAux1[j]:=Point[j];
          end;

          valueAux2:=FctAMinimiserMultiD(PointAux2);
          if (ValueAux2>=ValueAux1) or
             ((2*(ValueAux1-2*Value+ValueAux2)*sqr(ValueAux1-Value-BiggestDecrease)
             -BiggestDecrease*sqr(ValueAux1-ValueAux2))>=0)
          then {on ne change pas de direction }
          else begin
               LineMinimization(Point,DirectionVect,N,Value);
               for j:=1 to N do DirectionMat[j,IndiceBiggestDecrease]:=DirectionVect[j];
               end;
     end; {fin de la boucle en k }

end;

{ fin procedure minimisation multi D }


{ corps principal de MinimiserParaML }
begin
     { creation du vecteur a entrer dans la fct de minimisation }
     { L -> Point ; }
     TransfertLversP(L,AjusterDansL,P,N);
     MinimisationMultiD(N,Iter,tol,P,Value);
     TransfertPversL(P,L,AjusterDansL);
end;

end.
